@@ -27,7 +27,7 @@ from __future__ import print_function, unicode_literals
 
 import argparse
 import atexit
-from dataclasses import dataclass
+from dataclasses import dataclass, astuple
 from datetime import datetime, timedelta, timezone
 import logging
 import os.path
@@ -37,7 +37,17 @@ import shutil
 import subprocess
 import sys
 import tempfile
-from typing import Optional, Tuple, Type
+from typing import (
+    Callable,
+    Iterable,
+    Iterator,
+    Literal,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    TypeVar,
+)
 from urllib.parse import quote
 import yaml
 
@@ -54,7 +64,7 @@ except ImportError:
     #
     # couldn't trace it, so i silence it. original version has no type
     # hints at all, and is nearly identical.
-    def shlex_join(split_command):  # type: ignore[misc]
+    def shlex_join(split_command):  # type: ignore[misc,no-untyped-def]
         return " ".join(shlex.quote(arg) for arg in split_command)
 
 
@@ -65,10 +75,15 @@ try:
     import gi  # type: ignore
 
     gi.require_version("Gtk", "3.0")
-    from gi.repository import Gtk, Gdk, GdkPixbuf  # type: ignore
-except ImportError:
+    from gi.repository import GLib, Gtk, Gdk, GdkPixbuf  # type: ignore
+    from Gtk import MessageDialog  # type: ignore
+except (ImportError, ValueError):
+    # "ValueError('Namespace %s not available' % namespace)"
     gi = None
 
+    class MessageDialog:  # type: ignore
+        pass
+
 __epilog__ = """ This program will upload the provided files on the internet and
 notify the user when completed. The resulting URL will also be copied
 to the clipboard. Preprocessors can do fancy things before (only
@@ -82,7 +97,7 @@ class LoggingAction(argparse.Action):
     `basicConfig`.
     """
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
         """setup the action parameters
 
         This enforces a selection of logging levels. It also checks if
@@ -94,12 +109,14 @@ class LoggingAction(argparse.Action):
             kwargs["nargs"] = 0
         super().__init__(*args, **kwargs)
 
-    def __call__(self, parser, ns, values, option):
+    def __call__(self, parser, ns, values, option):  # type: ignore[no-untyped-def]
         """if const was specified it means argument-less parameters"""
         if self.const:
             logging.getLogger("").setLevel(self.const)
         else:
             logging.getLogger("").setLevel(values)
+        # cargo-culted from _StoreConstAction
+        setattr(ns, self.dest, self.const or values)
 
 
 class ConfigAction(argparse._StoreAction):
@@ -112,19 +129,19 @@ class ConfigAction(argparse._StoreAction):
     to fix that are welcome.
     """
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
         """the config action is a search path, so a list, so one or more argument"""
         kwargs["nargs"] = 1
         super().__init__(*args, **kwargs)
 
-    def __call__(self, parser, ns, values, option):
+    def __call__(self, parser, ns, values, option):  # type: ignore[no-untyped-def]
         """change defaults for the namespace, still allows overriding
         from commandline options"""
         for path in values:
             parser.set_defaults(**self.parse_config(path))
         super().__call__(parser, ns, values, option)
 
-    def parse_config(self, path):
+    def parse_config(self, path: str) -> dict:  # type: ignore[type-arg]
         """abstract implementation of config file parsing, should be overriden in subclasses"""
         raise NotImplementedError()
 
@@ -132,13 +149,13 @@ class ConfigAction(argparse._StoreAction):
 class YamlConfigAction(ConfigAction):
     """YAML config file parser action"""
 
-    def parse_config(self, path):
+    def parse_config(self, path: str) -> dict:  # type: ignore[type-arg]
         try:
             with open(os.path.expanduser(path), "r") as handle:
                 logging.debug("parsing path %s as YAML" % path)
                 return yaml.safe_load(handle) or {}
         except (FileNotFoundError, yaml.parser.ParserError) as e:
-            raise argparse.ArgumentError(self, e)
+            raise argparse.ArgumentError(self, str(e))
 
 
 class ConfigArgumentParser(argparse.ArgumentParser):
@@ -200,13 +217,13 @@ class ConfigArgumentParser(argparse.ArgumentParser):
 
     """
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
         super().__init__(*args, **kwargs)
         # a list of actions to fire with their defaults if not fired
         # during parsing
         self._delayed_config_action = []
 
-    def _add_action(self, action):
+    def _add_action(self, action: argparse.Action) -> argparse.Action:  # type: ignore[override]
         # this overrides the add_argument() routine, which is where
         # actions get registered in the argparse module.
         #
@@ -221,8 +238,13 @@ class ConfigArgumentParser(argparse.ArgumentParser):
         action = super()._add_action(action)
         if isinstance(action, ConfigAction) and action.default is not None:
             self._delayed_config_action.append(action)
+        return action
 
-    def parse_args(self, args=None, namespace=None):
+    def parse_args(  # type: ignore[override]
+        self,
+        args: Optional[Sequence[str]] = None,
+        namespace: Optional[argparse.Namespace] = None,
+    ) -> argparse.Namespace:
         # we do a first failsafe pass on the commandline to find out
         # if we have any "config" parameters specified, in which case
         # we must *not* load the default config file
@@ -255,7 +277,7 @@ class ConfigArgumentParser(argparse.ArgumentParser):
         # note that this will load the config file a second time
         return super().parse_args(args, namespace)
 
-    def default_config(self):
+    def default_config(self) -> Iterable[str]:
         """handy shortcut to detect commonly used config paths"""
         return [
             os.path.join(
@@ -265,11 +287,11 @@ class ConfigArgumentParser(argparse.ArgumentParser):
 
 
 class PubpasteArgumentParser(ConfigArgumentParser):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs):  # type: ignore
         """
         override constructor to setup our arguments and config file
         """
-        super().__init__(description=__doc__, epilog=__epilog__, *args, **kwargs)
+        super().__init__(description=__doc__, epilog=__epilog__, *args, **kwargs)  # type: ignore
         # This is the usage of weasel's publish:
         #
         # usage: publish [<src> [<src> ...]]
@@ -317,7 +339,9 @@ class PubpasteArgumentParser(ConfigArgumentParser):
             help="how long to keep the entry, in days, default: %(default)s",
         )
         group = group.add_mutually_exclusive_group()
-        group.add_argument("-T", "--token", help="secret token, default: generated")
+        group.add_argument(
+            "-T", "--token", type=str, help="secret token, default: generated"
+        )
         group.add_argument(
             "-R", "--republish", action="store_true", help="reuse previous secret token"
         )
@@ -336,26 +360,32 @@ class PubpasteArgumentParser(ConfigArgumentParser):
             "--stdin-name",
             metavar="NAME",
             default="stdin.txt",
+            type=str,
             help="use NAME as filename when reading from stdin, default: %(default)s",
         )
         self.add_argument(
             "-o",
             "--output",
+            type=str,
             help=argparse.SUPPRESS,  # only in configuration file
             default=None,
         )
         self.add_argument(
             "--url-prefix",
+            type=str,
             help=argparse.SUPPRESS,  # only in configuration file
             default=None,
         )
         self.add_argument(
             "--base-dir",
+            type=str,
             help=argparse.SUPPRESS,  # only in configuration file
             default=None,
         )
         self.add_argument(
-            "--save-screenshots", help=argparse.SUPPRESS  # only in configuration file
+            "--save-screenshots",
+            type=str,
+            help=argparse.SUPPRESS,  # only in configuration file
         )
         self.add_argument(
             "--select",
@@ -440,7 +470,7 @@ class PubpasteArgumentParser(ConfigArgumentParser):
         )
         self.add_argument("files", nargs="*", help="files to upload")
 
-    def parse_args(self, *args, **kwargs):
+    def parse_args(self, *args, **kwargs):  # type: ignore[no-untyped-def]
         res = super().parse_args(*args, **kwargs)
         if res.output and not res.output.endswith("/"):
             res.output += "/"
@@ -455,7 +485,7 @@ class PubpasteArgumentParser(ConfigArgumentParser):
             self.error(
                 "command %s and files cannot be specified together" % res.command
             )
-        if res.command and res.command == "purge" and args.base_dir is None:
+        if res.command and res.command == "purge" and res.base_dir is None:
             self.error("no --base-dir specified, cannot find files to purge")
         if res.command and res.command in ("xselection", "xclipboard") and res.relative:
             # this is to workaround a bug where the tmpfile path gets
@@ -467,11 +497,11 @@ class PubpasteArgumentParser(ConfigArgumentParser):
 
 
 class Processor(object):
-    def __init__(self, args, dryrun=False):
+    def __init__(self, args: argparse.Namespace, dryrun: bool = False) -> None:
         self.args = args
         self.dryrun = dryrun
 
-    def process(self, paths, tmp_dir):
+    def process(self, paths: Iterable[Path], tmp_dir: str) -> Iterable[str]:
         """process the given paths
 
         Returns the modified paths, or None if not modified. A tmp_dir
@@ -486,7 +516,7 @@ class Sigal(Processor):
     directory. it returns only the "build" directory.
     """
 
-    def process(self, paths, tmp_dir):
+    def process(self, paths: Iterable[Path], tmp_dir: str) -> Iterable[str]:
         output_dir = Path(tmp_dir) / Path(self.args.token)
         if not output_dir.exists() and not self.dryrun:
             logging.info("creating directory %s", output_dir)
@@ -529,7 +559,7 @@ class Sigal(Processor):
         return (str(build_dir),)
 
     @classmethod
-    def sigal_minimal_config(cls):
+    def sigal_minimal_config(cls) -> str:
         """a string representing a good minimal sigal configuration for our
         use case.
 
@@ -553,44 +583,209 @@ thumb_video_delay = '3'
 """
 
 
-class Maim(Processor):
-    """The Main processor will take a screenshot and return the file after
-    copying it to the temporary directory"""
+class TimerWindow(MessageDialog):  # type: ignore[misc]
+    """This widget will show a timer and a spinner in a window for the
+    given delay, then close the dialog.
 
-    def __init__(self, args, dryrun=False, save_screenshots=False):
-        super().__init__(args, dryrun=dryrun)
-        if args.select:
-            self.command = "maim --select --delay=3 '%s'"
+    It's designed to be called with `run()`."""
+
+    DELAY = 3  # how long to wait, in seconds
+
+    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
+        super().__init__(title="Taking screenshot...")
+
+        mainBox = self.get_content_area()
+        self.spinner = Gtk.Spinner()
+        mainBox.pack_start(self.spinner, True, True, 0)
+
+        self.timeout_id = None
+        self.connect("destroy", self.clean_timers)
+        self.start_timer()
+
+    def clean_timers(self, *args) -> None:  # type: ignore[no-untyped-def]
+        """Ensure the timeout function is stopped"""
+        if self.timeout_id:
+            GLib.source_remove(self.timeout_id)
+            self.timeout_id = None
+
+    def on_timeout(self, *args, **kwargs) -> bool:  # type: ignore[no-untyped-def]
+        """A timeout function.
+
+        This is not a precise timer since next timeout
+        is recalculated based on the current time.
+
+        When this returns False, the timer is stopped.x"""
+        self.counter -= 1
+        if self.counter <= 0:
+            self.clean_timers()
+            self.spinner.stop()
+            self.response(False)
+            return False
+        elif self.counter < 4:
+            self.format_secondary_text("Say cheese!")
         else:
-            self.command = "maim --delay=3 '%s'"
+            self.format_secondary_text(
+                "Taking screenshot in  " + str(int(self.counter / 4)) + " seconds!"
+            )
+        return True
+
+    def start_timer(self) -> None:
+        """Start the timer
+
+        time out will check every 250 milliseconds (1/4 of a second),
+        so that the display is *approximately* accurate.
+
+        Since we're not telling the time here, it shouldn't matter
+        *too* much, but it looks better than rounded seconds,
+        presumably.
+        """
+        self.counter = 4 * self.DELAY
+        self.format_secondary_text(
+            "Taking screenshot in  " + str(int(self.counter / 4)) + " seconds!"
+        )
+        self.spinner.start()
+        self.timeout_id = GLib.timeout_add(250, self.on_timeout, None)
+
+
+class ScreenshotProcessor(Processor):
+    def __init__(
+        self,
+        args: argparse.Namespace,
+        dryrun: bool = False,
+        save_screenshots: Optional[str] = None,
+    ) -> None:
+        super().__init__(args, dryrun=dryrun)
+        self.select = args.select
         self.save_screenshots = save_screenshots
 
-    def process(self, paths, tmp_dir):
-        """wrap main around a timer, preview and prompt"""
+    def _select_paths(self, tmp_dir: str) -> Tuple[Path, Path]:
+        # build tempdir or use the assigned screenshot dir
         if self.save_screenshots:
             snaps_dir = Path(os.path.expanduser(self.save_screenshots))
             if not snaps_dir.exists():
                 snaps_dir.mkdir()
         else:
             snaps_dir = Path(tmp_dir)
+        # set a the screenshot filename
         snap_basename = Path(
             "snap-" + datetime.now().strftime("%Y%m%dT%H%M%S%Z") + ".png"
         )
+        target = Path(tmp_dir) / snap_basename
         snap_path = snaps_dir / snap_basename
-        # XXX: this is just horrible. can't the screenshot tool do the right thing here?
-        #
-        # TODO: at the very least, make our own popup, because right
-        # now this doesn't catch errors from maim, as errors from the
-        # subprocess are not trickled back up through xterm
-        command = (
-            "xterm",
-            "-title",
-            "pop-up",
-            "-geometry",
-            "80x3+5+5",
-            "-e",
-            self.command % snap_path,
-        )
+        return snap_path, target
+
+    def gtk_wait_for_timeout(self) -> None:
+        timer = TimerWindow()
+        timer.show_all()
+        timer.run()
+        timer.destroy()
+        while Gtk.events_pending():
+            Gtk.main_iteration()
+
+    def gtk_confirm(self) -> bool:
+        mb = Gtk.MessageDialog()
+        mb.add_button("Yes", 0)
+        mb.add_button("No", 1)
+        mb.set_title("Upload screenshot publicly?")
+        mb.set_markup("Should we upload this screenshot publicly now?")
+        ret = bool(mb.run() == 0)
+        mb.destroy()
+        return ret
+
+    def confirm_upload(self, snap_path: Path, target: Path) -> Iterable[str]:
+        if self.gtk_confirm():
+            if not self.save_screenshots:
+                logging.debug("returning %s", target)
+                return (str(target),)
+            shutil.copy(str(snap_path), target)
+            # pass the resulting file back to the uploader
+            logging.info("screenshot copied from %s to %s", snap_path, target)
+            return (str(target),)
+        else:
+            logging.info("user declined, aborting")
+        return []
+
+
+class Grim(ScreenshotProcessor):
+    """The Grim processor will take a screenshot and return the file after
+    copying it into a temporary directory.
+
+    An alternative to this is shotman, not in Debian:
+
+    https://git.sr.ht/~whynothugo/shotman"""
+
+    def process(self, paths: Iterable[Path], tmp_dir: str) -> Iterable[str]:
+        """wrap grim around a timer, preview and prompt"""
+        snap_path, target = self._select_paths(tmp_dir)
+        if self.select:
+            logging.debug("selecting area with slurp")
+            select_pipe = subprocess.run(["slurp"], stdout=subprocess.PIPE, check=False)
+            if select_pipe.returncode != 0:
+                notify_user("selection aborted")
+                return []
+            command = [
+                "grim",
+                "-g",
+                select_pipe.stdout.strip().decode("ascii"),
+                str(snap_path),
+            ]
+        else:
+            command = ["grim", str(snap_path)]
+
+        self.gtk_wait_for_timeout()
+
+        if self.dryrun:
+            logging.warning("running in dry run, not taking screenshot")
+            return []
+
+        logging.debug("running screenshot with '%s'", shlex_join(command))
+        success = subprocess.call(command) == 0
+        if not success:
+            notify_user("screenshot command failed, aborting")
+            return []
+        if not snap_path.exists():
+            notify_user("snapshot file missing, is maim installed?")
+            return []
+        if snap_path.stat().st_size <= 0:
+            notify_user("snapshot file is empty, aborting")
+            snap_path.unlink()
+            return []
+        # we delegate the customization of this to XDG
+        command = ["xdg-open", str(snap_path)]
+        logging.debug("opening image viewer with %s", command)
+        subprocess.Popen(command)  # nowait
+        return self.confirm_upload(snap_path, target)
+
+
+class Maim(ScreenshotProcessor):
+    """The Main processor will take a screenshot and return the file after
+    copying it to the temporary directory"""
+
+    def __init__(
+        self,
+        args: argparse.Namespace,
+        dryrun: bool = False,
+        save_screenshots: Optional[str] = None,
+    ) -> None:
+        super().__init__(args, dryrun=dryrun)
+        self.save_screenshots = save_screenshots
+
+    def process(self, paths: Iterable[Path], tmp_dir: str) -> Iterable[str]:
+        """wrap main around a timer, preview and prompt"""
+        if self.select:
+            logging.debug("selecting area with slop")
+            select_pipe = subprocess.run(["slop"], stdout=subprocess.PIPE, check=False)
+            if select_pipe.returncode != 0:
+                notify_user("selection aborted")
+                return []
+        self.gtk_wait_for_timeout()
+        snap_path, target = self._select_paths(tmp_dir)
+        command = [
+            "maim",
+            "-g",
+            select_pipe.stdout.strip().decode("ascii"),
+            str(snap_path),
+        ]
         logging.debug("running screenshot with %s", shlex_join(command))
         if not self.dryrun:
             success = subprocess.call(command) == 0
@@ -605,49 +800,29 @@ class Maim(Processor):
             snap_path.unlink()
             return []
         # we delegate the customization of this to XDG
-        subprocess.Popen(("xdg-open", str(snap_path)))  # nowait
-        # XXX: gah. we *could* just do GTK here?
-        command = (
-            "xmessage",
-            "-buttons",
-            "okay:0,cancel:1",
-            "-center",
-            "-default",
-            "okay",
-            "upload snapshot publically?",
-        )
-        try:
-            retcode = subprocess.call(command)
-        except OSError as e:
-            logging.error("cannot find xmessage? %s", e)
-            return []
-        if retcode == 0:
-            target = str(Path(tmp_dir) / snap_basename)
-            if not self.save_screenshots:
-                logging.debug("returning %s", tmp_dir)
-                return (target,)
-            shutil.copy(str(snap_path), target)
-            # pass the resulting file back to the uploader
-            logging.info("screenshot copied from %s to %s", snap_path, target)
-            return (target,)
-        elif retcode == 1:
-            logging.info("user declined, aborting")
-        elif retcode > 1:
-            logging.error("unexpected error from xmessage")
-        elif retcode < 0:
-            logging.error("xmessage terminated by signal %s", -retcode)
-        return []
+        command = ["xdg-open", str(snap_path)]
+        logging.debug("opening image viewer with %s", command)
+        subprocess.Popen(command)  # nowait
+        return self.confirm_upload(snap_path, target)
 
 
 class Uploader(object):
     """an abstract class to wrap around upload objects"""
 
-    def __init__(self, target, dryrun=False, follow_symlinks=False):
+    def __init__(
+        self, target: str, dryrun: bool = False, follow_symlinks: bool = False
+    ) -> None:
         self.target = target
         self.dryrun = dryrun
         self.follow_symlinks = follow_symlinks
 
-    def upload(self, paths, target=None, single=False):
+    def upload(
+        self,
+        path: str,
+        target: Optional[str] = None,
+        single: Optional[bool] = False,
+        cwd: Optional[str] = None,
+    ) -> bool:
         """upload the given path to the target
 
         "Single" is an indication by the caller of whether or not this
@@ -657,7 +832,7 @@ class Uploader(object):
 
 
 class RsyncUploader(Uploader):
-    base_rsync_command = (
+    base_rsync_command: Tuple[str, ...] = (
         "rsync",
         "--recursive",
         "--compress",
@@ -665,7 +840,13 @@ class RsyncUploader(Uploader):
         "--chmod=u=rwX,go=rX",
     )
 
-    def upload(self, path, target=None, single=False, cwd=None):
+    def upload(
+        self,
+        path: str,
+        target: Optional[str] = None,
+        single: Optional[bool] = False,
+        cwd: Optional[str] = None,
+    ) -> bool:
         """upload the given path to the target
 
         "Single" is an indication by the caller of whether or not this
@@ -744,7 +925,7 @@ class RsyncUploader(Uploader):
             return self.dryrun
         return subprocess.call(command, cwd=cwd) == 0
 
-    def delete(self, target=None):
+    def delete(self, target: Optional[str] = None) -> bool:
         """delete the given target, or the one from the constructor
 
         This is done by synchronizing it with an empty temporary
@@ -766,55 +947,65 @@ class RsyncUploader(Uploader):
             return subprocess.check_call(command) == 0
 
 
+# we can't refer to History.Entry in History.Entry._make, see:
+#
+# https://stackoverflow.com/a/54534451/1174784
+#
+# this is fixed in PEP 673 (Python 3.11): https://peps.python.org/pep-0673/
+EntryType = TypeVar("EntryType", bound="History.Entry")
+
+
 class History:
     TTL_PATH = ".publish.ttl"
 
     @dataclass
     class Entry:
-        date: Optional[str] = None
-        time: Optional[str] = None
-        token: Optional[str] = None
+        date: str
+        time: str
+        token: str
         uri: Optional[str] = None
 
-        @classmethod
-        def _make(cls, line):
+        @staticmethod
+        # we need to quote the type here because History.Entry is not defined yet
+        # and yep, that works.
+        def _make(line: str) -> "History.Entry":
             items = line.strip().split(" ", 4)
             return History.Entry(*items)
 
-        def __str__(self):
-            return " ".join([self.date, self.time, self.token, self.uri])
+        def __str__(self) -> str:
+            return " ".join(astuple(self))
 
-    def __init__(self, path: str = None):
+    def __init__(self, path: Optional[str] = None) -> None:
         if path is None:
             path = os.path.expanduser("~/.publish.history")
         self.path = path
         self.recorded = False
 
-    def append(self, token, uri):
+    def append(self, token: str, uri: Optional[str]) -> None:
         timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S%Z")
         with open(self.path, "a") as fp:
-            fp.write(" ".join((timestamp, token, uri)) + "\n")
+            fp.write(" ".join((timestamp, token, uri or "")) + "\n")
 
-    def append_once(self, token, uri):
+    def append_once(self, token: str, uri: Optional[str]) -> None:
         if not self.recorded:
             self.append(token, uri)
             self.recorded = True
 
-    def remove(self, tokens=()):
+    def remove(self, tokens: Iterable[str] = ()) -> None:
         """remove the given tokens from history"""
         with open(self.path, "r") as old, open(self.path + ".new", "w") as new:
-            for line in old.readlines():
-                entry = History.Entry._make(line)
+            for line in old:
+                entry: History.Entry = History.Entry._make(line)
                 if entry.token not in tokens:
                     new.write(line)
         os.rename(self.path + ".new", self.path)
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Entry]:
         with open(self.path, "r") as fp:
-            for line in fp.readlines():
+            for line in fp:
                 yield History.Entry._make(line)
 
-    def last(self):
+    def last(self) -> Optional[Entry]:
         """find last history token
 
         This is separate from the above generator to avoid messing
@@ -823,14 +1014,14 @@ class History:
         # date, time, token, uri
         entry = None
         with open(self.path, "r") as fp:
-            for line in fp.readlines():
+            for line in fp:
                 pass
         if line:
             entry = History.Entry._make(line)
             logging.debug("found last entry: %s", entry)
         return entry
 
-    def purge(self, base_dir):
+    def purge(self, base_dir: str) -> None:
         logging.debug("purging from %s", base_dir)
         for path in [p for p in Path(base_dir).iterdir() if p.is_dir()]:
             token = path.name
@@ -876,25 +1067,28 @@ class History:
                 )
 
 
-def secret_token():
+def secret_token() -> Tuple[str, str]:
     """return a secret token made of multiple components, as a tuple"""
     return datetime.now().strftime("%Y-%m-%d"), secrets.token_urlsafe()
 
 
-class AbstractClipboard:
-    def __init__(self, selection: str):
+class BaseClipboard:
+    def __init__(self, selection: Literal["CLIPBOARD", "PRIMARY"]) -> None:
         """configure the clipboard, based on whether the clipboard is PRIMARY
         or CLIPBOARD
         """
+        self.selection = selection
+
+    def __str__(self) -> str:
+        return "%s(%s)" % (type(self).__name__, self.selection)
+
+    def put_text(self, text: str) -> bool:
         raise NotImplementedError()
 
-    def put_text(self, text: str):
+    def put_image(self, path: bytes) -> bool:
         raise NotImplementedError()
 
-    def put_image(self, path: bytes):
-        raise NotImplementedError()
-
-    def get(self) -> Tuple[Optional[bytes], Optional[str]]:
+    def get(self) -> Tuple[Optional[bytes], Optional[Literal["png", "jpg", "txt"]]]:
         raise NotImplementedError()
 
 
@@ -906,10 +1100,10 @@ class AbstractClipboard:
 # also, pyperclip actually *fails* to use gi in my tests, in Debian
 # bullseye, so really, it's a lot of code for nothing for us. Worse,
 # it doesn't support copying images the way we do.
-class GtkClipboard(AbstractClipboard):
+class GtkClipboard(BaseClipboard):
     "Clipboard implementation with a GTK/GI backend"
 
-    def __init__(self, selection):
+    def __init__(self, selection: Literal["CLIPBOARD", "PRIMARY"]) -> None:
         """initialize a clipboard based on CLIPBOARD or PRIMARY selection
 
         we use a string argument here to abstract away the GTK stuff"""
@@ -919,8 +1113,9 @@ class GtkClipboard(AbstractClipboard):
         elif selection == "PRIMARY":
             self._cb = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
         logging.debug("created clipboard with %s selection", selection)
+        super().__init__(selection)
 
-    def _gdk_event_handler(self, event, user_data=None):
+    def _gdk_event_handler(self, event, user_data=None):  # type: ignore[no-untyped-def]
         logging.debug("event: %r, user_data: %r", event, user_data)
         # pass the event to GTK so we actually send the paste back,
         # and do whatever it is GTK does with events
@@ -933,7 +1128,7 @@ class GtkClipboard(AbstractClipboard):
             logging.info("selection picked up by another process, exiting")
             Gtk.main_quit()
 
-    def _wait_for_paste(self):
+    def _wait_for_paste(self) -> None:
         # add an event handler to exit early when something is pasted
         # https://developer.gnome.org/gdk3/stable/gdk3-Events.html#gdk-event-handler-set
         Gdk.event_handler_set(self._gdk_event_handler)
@@ -941,10 +1136,10 @@ class GtkClipboard(AbstractClipboard):
         # Gtk.main(). Believe me, I tried.
         Gtk.main()
 
-    def put_text(self, text: str):
+    def put_text(self, text: str) -> bool:
         """set the clipboard to the given text"""
 
-        def callback():
+        def callback() -> None:
             logging.debug("setting clipboard text to %r", text)
             # NOTE: yes, it's silly that set_text() asks for the data
             # length, this is python after all?
@@ -952,17 +1147,17 @@ class GtkClipboard(AbstractClipboard):
 
         return self._put(callback)
 
-    def put_image(self, path: bytes):
+    def put_image(self, path: bytes) -> bool:
         """load the image from the given path name and put in the clipboard"""
 
-        def callback():
+        def callback() -> None:
             pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
             logging.debug("setting clipboard image to %r", pixbuf)
             self._cb.set_image(pixbuf)
 
         return self._put(callback)
 
-    def _put(self, callback):
+    def _put(self, callback: Callable[[], None]) -> bool:
         """utility to do the actual paste, fork and wait for the paste to complete
 
         This will just call the given callback, which is assumed to
@@ -986,7 +1181,7 @@ class GtkClipboard(AbstractClipboard):
             # we use _exit() to avoid firing atexit() in double
             os._exit(0)
 
-    def get(self):
+    def get(self) -> Tuple[Optional[bytes], Optional[Literal["png", "jpg", "txt"]]]:
         """get the clipboard content, possible as an image or, failing that, as text
 
         GTK clipboards also support "rich text" and "URI", but I'm not
@@ -1024,19 +1219,71 @@ class GtkClipboard(AbstractClipboard):
             return None, None
 
 
+class WaylandClipboard(BaseClipboard):
+    def __init__(self, selection: Literal["CLIPBOARD", "PRIMARY"]) -> None:
+        self.extra = []
+        if selection == "PRIMARY":
+            self.extra = ["--primary"]
+        super().__init__(selection)
+
+    def put_text(self, text: str) -> bool:
+        return self._put_data(text.encode("utf-8"))
+
+    def put_image(self, data: bytes) -> bool:
+        return self._put_data(data)
+
+    def _put_data(self, data: bytes) -> bool:
+        command = ["wl-copy"] + self.extra
+        logging.debug("sending %d bytes to %s", len(data), command)
+        p = subprocess.Popen(command, stdin=subprocess.PIPE)
+        p.communicate(data)
+        if p.returncode == 0:
+            return True
+        else:
+            logging.warning("could not copy to clipboard with: %r", command)
+            return False
+
+    def get(self) -> Tuple[Optional[bytes], Optional[Literal["png", "jpg", "txt"]]]:
+        command = ["wl-paste", "--list-types"] + self.extra
+        logging.debug("listing supported types with %s", command)
+        mimetype_output = subprocess.run(command, stdout=subprocess.PIPE)
+        mimetypes = mimetype_output.stdout.decode("ascii").split("\n")
+        extension: Optional[Literal["png", "jpg", "txt"]] = None
+        if mimetypes == ["No selection"]:
+            return None, None
+        if "image/png" in mimetypes:
+            extension = "png"
+            mimetype = "image/png"
+        elif "image/jpeg" in mimetypes:
+            extension = "jpg"
+            mimetype = "image/jpeg"
+        elif "text/plain" in mimetypes:
+            extension = "txt"
+            mimetype = "text/plain"
+        else:
+            logging.warning("unsupported clipboard mimetypes: %s", mimetypes)
+            return None, None
+        command = ["wl-paste", "--type", mimetype] + self.extra
+        logging.debug("extracting data with %s", command)
+        data = subprocess.run(command, stdout=subprocess.PIPE).stdout
+        return data, extension
+
+
 # TODO: xsel support? might be hard because it relies on a clipboard
 # manager (e.g. xclipboard(1)) to store the selection for us, which we
 # can't rely on... The equivalent in GTK is the `store()` command,
 # which didn't work in my environment (i3 minimal desktop).
-class XclipClipboard(AbstractClipboard):
-    def __init__(self, selection):
+class XclipClipboard(BaseClipboard):
+    def __init__(self, selection: Literal["CLIPBOARD", "PRIMARY"]) -> None:
         assert selection in ("CLIPBOARD", "PRIMARY"), "invalid clipboard selection used"
         self.command = ("xclip", "-selection", selection.lower())
+        super().__init__(selection)
 
-    def put_text(self, text: str):
+    def put_text(self, text: str) -> bool:
         if not os.environ.get("DISPLAY"):
             logging.warning("could not copy to clipboard without a DISPLAY variable")
             return False
+        logging.debug("copying clipboard with: %s", self.command)
         p = subprocess.Popen(self.command, stdin=subprocess.PIPE)
         p.communicate(text.encode("utf-8"))
         if p.returncode == 0:
@@ -1045,11 +1292,12 @@ class XclipClipboard(AbstractClipboard):
             logging.warning("could not copy to clipboard with: %r", self.command)
             return False
 
-    def put_image(self, data: bytes):
+    def put_image(self, data: bytes) -> bool:
         raise NotImplementedError("xclip does not support copying images")
 
-    def get(self):
+    def get(self) -> Tuple[Optional[bytes], Optional[Literal["png", "txt"]]]:
         command = list(self.command) + ["-o"]
+        logging.debug("pasting clipboard with: %s", self.command)
         try:
             clipboard = subprocess.check_output(command)
         except (FileNotFoundError, subprocess.CalledProcessError) as e:
@@ -1058,23 +1306,33 @@ class XclipClipboard(AbstractClipboard):
         return clipboard, "txt"
 
 
-if gi:
-    Clipboard: Type[AbstractClipboard] = GtkClipboard
+def in_wayland() -> bool:
+    return bool(
+        os.environ.get("XDG_SESSION_TYPE", "xorg") == "wayland"
+        or os.environ.get("WAYLAND_DISPLAY")
+    )
+
+
+if in_wayland():
+    Clipboard: Type[BaseClipboard] = WaylandClipboard
+elif gi:
+    Clipboard = GtkClipboard
 else:
     Clipboard = XclipClipboard
 
 
-def notify_user(message):
+def notify_user(message: str) -> None:
     logging.warning(message)
     if os.environ.get("DISPLAY") is None:
         return
     command = ("notify-send", message)
     # do not fail on notifications
+    logging.debug("running %s", command)
     if subprocess.call(command) != 0:
         logging.warning("failed to run command %r", command)
 
 
-def main():
+def main() -> None:
     """
     This program does the following:
 
@@ -1088,7 +1346,7 @@ def main():
     8. notify the user, probably with notify-send
     """
     logging.basicConfig(format="%(message)s", level="WARNING")
-    args = PubpasteArgumentParser().parse_args()
+    args = PubpasteArgumentParser().parse_args()  # type: ignore[no-untyped-call]
 
     history = History()
 
@@ -1169,10 +1427,10 @@ def main():
         sys.exit(0)
 
     elif args.command in ("xselection", "xclipboard"):
-        logging.info("reading from xselection...")
         clipboard_primary = Clipboard(
             "PRIMARY" if args.command == "xselection" else "CLIPBOARD"
         )
+        logging.info("reading from xselection with %s...", clipboard_primary)
         data, extension = clipboard_primary.get()
         if data is None:
             logging.error("no clipboard data found, aborting")
@@ -1190,8 +1448,16 @@ def main():
             dump_content = [str(clipboard_tmp_path)]
 
     elif args.command == "screenshot":
-        m = Maim(args, dryrun=args.dryrun, save_screenshots=args.save_screenshots)
-        args.files = m.process([], tmp_dir.name)
+        if in_wayland():
+            logging.info("running under Wayland, trying screenshot with Grim")
+            args.files = Grim(
+                args, dryrun=args.dryrun, save_screenshots=args.save_screenshots
+            ).process([], tmp_dir.name)
+        else:  # Xorg case
+            logging.info("running under Xorg, trying screenshot with Maim")
+            args.files = Maim(
+                args, dryrun=args.dryrun, save_screenshots=args.save_screenshots
+            ).process([], tmp_dir.name)
 
     elif args.command == "gallery":
         logging.debug("processing files %s with Sigal", args.files)
@@ -1222,8 +1488,7 @@ def main():
         with tempfile.TemporaryDirectory() as ttl_tmp_dir:
             ttl_file = ttl_tmp_dir + "/" + history.TTL_PATH
             if not args.dryrun:
-                with open(ttl_file, "w") as fp:
-                    fp.write(args.ttl + "\n")
+                Path(ttl_file).write_text(args.ttl + "\n")
             if not uploader.upload(history.TTL_PATH, cwd=ttl_tmp_dir):
                 notify_user("failed to upload TTL file for %s" % args.token)
 
